#  Copyright 2017, Iain Dunning, Joey Huchette, Miles Lubin, and contributors
#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
#############################################################################
# JuMP
# An algebraic modeling language for Julia
# See https://github.com/jump-dev/JuMP.jl
#############################################################################

module TestObjective

using JuMP
using Test

function runtests()
    for name in names(@__MODULE__; all = true)
        if startswith("$(name)", "test_")
            @testset "$(name)" begin
                getfield(@__MODULE__, name)()
            end
        end
    end
    return
end

include(joinpath(@__DIR__, "JuMPExtension.jl"))

struct DummyOptimizer <: MOI.AbstractOptimizer end

MOI.is_empty(::DummyOptimizer) = true

function test_unsupported_objective_function()
    model = Model(DummyOptimizer)
    func = MOI.VariableIndex(1)
    @test_throws ErrorException JuMP.set_objective_function(model, func)
    return
end

function test_unsupported_function_in_macro()
    model = Model()
    @variable(model, x[1:2])
    exception = ErrorException(
        "The objective function `$VariableRef[x[1]," *
        " x[2]]` is not supported by JuMP.",
    )
    @test_throws exception @objective(model, Min, x)
    return
end

function objectives_test(
    ModelType::Type{<:JuMP.AbstractModel},
    VariableRefType::Type{<:JuMP.AbstractVariableRef},
)
    AffExprType = JuMP.GenericAffExpr{Float64,VariableRefType}
    QuadExprType = JuMP.GenericQuadExpr{Float64,VariableRefType}

    @testset "objective_sense set and get" begin
        model = ModelType()
        JuMP.set_objective_sense(model, FEASIBILITY_SENSE)
        @test FEASIBILITY_SENSE == @inferred JuMP.objective_sense(model)
    end

    @testset "SingleVariable objectives" begin
        m = ModelType()
        @variable(m, x)

        @objective(m, Min, x)
        @test MIN_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.objective_function_type(m) == VariableRefType
        @test JuMP.objective_function(m) == x
        @test x == @inferred JuMP.objective_function(m, VariableRefType)

        @objective(m, Max, x)
        @test MAX_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.objective_function_type(m) == VariableRefType
        @test JuMP.objective_function(m) == x
        @test x == @inferred JuMP.objective_function(m, VariableRefType)
    end

    @testset "Linear objectives" begin
        m = ModelType()
        @variable(m, x)

        @objective(m, Min, 2x)
        @test MIN_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.objective_function_type(m) == AffExprType
        @test JuMP.isequal_canonical(JuMP.objective_function(m), 2x)
        @test JuMP.isequal_canonical(
            2x,
            @inferred JuMP.objective_function(m, AffExprType)
        )

        @objective(m, Max, x + 3x + 1)
        @test MAX_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.objective_function_type(m) == AffExprType
        @test JuMP.isequal_canonical(JuMP.objective_function(m), 4x + 1)
        @test JuMP.isequal_canonical(
            4x + 1,
            @inferred JuMP.objective_function(m, AffExprType)
        )
    end

    @testset "Quadratic objectives" begin
        m = ModelType()
        @variable(m, x)

        @objective(m, Min, x^2 + 2x)
        @test MIN_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.objective_function_type(m) == QuadExprType
        @test JuMP.isequal_canonical(JuMP.objective_function(m), x^2 + 2x)
        @test JuMP.isequal_canonical(
            x^2 + 2x,
            @inferred JuMP.objective_function(m, QuadExprType)
        )
        @test_throws InexactError JuMP.objective_function(m, AffExprType)
    end

    @testset "Sense as symbol" begin
        m = ModelType()
        @variable(m, x)

        @test_throws ErrorException @objective(m, :Min, 2x)
    end

    @testset "Sense in variable" begin
        m = ModelType()
        @variable(m, x)

        sense = MIN_SENSE
        @objective(m, sense, 2x)
        @test MIN_SENSE == @inferred JuMP.objective_sense(m)
        @test JuMP.isequal_canonical(
            2x,
            @inferred JuMP.objective_function(m, AffExprType)
        )

        sense = :Min
        @test_throws ErrorException @objective(m, sense, 2x)
    end

    @testset "Constant objective" begin
        model = ModelType()
        @objective(model, Min, 3)
        @test JuMP.objective_sense(model) == MIN_SENSE
        @test JuMP.isequal_canonical(
            AffExprType(3.0),
            JuMP.objective_function(model, AffExprType),
        )
    end
end

function objective_coeff_update_test(
    ModelType::Type{<:JuMP.AbstractModel},
    VariableRefType::Type{<:JuMP.AbstractVariableRef},
)
    @testset "Linear objective changes" begin
        m = ModelType()
        @variable(m, x)

        @objective(m, Max, x)
        set_objective_coefficient(m, x, 4.0)
        @test JuMP.isequal_canonical(JuMP.objective_function(m), 4x)

        @variable(m, y)
        @objective(m, Max, x + y)
        set_objective_coefficient(m, x, 4.0)
        @test JuMP.isequal_canonical(JuMP.objective_function(m), 4x + y)

        @objective(m, Min, x)
        set_objective_coefficient(m, y, 2.0)
        @test JuMP.isequal_canonical(JuMP.objective_function(m), x + 2.0 * y)
    end

    @testset "Quadratic objective changes" begin
        m = ModelType()
        @variable(m, x)

        @objective(m, Max, x^2 + x)
        set_objective_coefficient(m, x, 4.0)
        @test JuMP.isequal_canonical(JuMP.objective_function(m), x^2 + 4x)
    end
end

@testset "Objectives for JuMP.Model" begin
    objectives_test(Model, VariableRef)
    objective_coeff_update_test(Model, VariableRef)
end

@testset "Objectives for JuMPExtension.MyModel" begin
    objectives_test(JuMPExtension.MyModel, JuMPExtension.MyVariableRef)
end

end  # module

TestObjective.runtests()
